/* $Id: ComputerPlayer.java,v 1.15 2007/11/28 17:49:42 ekstrand Exp $ */

package csci1902.hw6.ai;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.*;
import java.util.concurrent.*;

import csci1902.hw6.game.CellPosition;
import csci1902.hw6.game.Game;
import csci1902.hw6.game.MoveListener;
import csci1902.hw6.game.Player;
import csci1902.hw6.game.PlayerID;
import static javax.swing.SwingUtilities.invokeLater;

/**
 * Class implementing a computer AI player.  This class is a {@link Player},
 * so the game can work with it and let it play Tic-Tac-Toe.  It uses an
 * {@link AIEngine} to implement the artificial intellence, and implements
 * some of the behind-the-scenes bookeeping needed to keep the AI moving
 * through the game.
 *
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 */
public class ComputerPlayer extends Player {
    /** The depth cutoff for this player's AI */       
    private int depthCutOff;
    /** The heuristic AI's are supposed to use */
    private BoardHeuristic heuristic;
    /** The thread in which the AI runs. */
    private AIThread aiThread;

    /** Logger for outputting debug messages. */
    private static Logger logger =
        Logger.getLogger("csci1902.hw6.ComputerPlayer");

    /**
     * Construct a new computer player.
     * @param heuristic The scoring heuristic for the AI to use.
     * @param player The player the computer will play as.
     * @param depthCutOff The maximum depth of the AI.
     */
    public ComputerPlayer(BoardHeuristic heuristic,
                          PlayerID player, int depthCutOff) {
        super(player);
        this.heuristic = heuristic;
        this.depthCutOff = depthCutOff;
        aiThread = null;
    }

    @Override
    public void setGame(Game game) {
        super.setGame(game);
        
        if (aiThread != null)
            aiThread.shutdown();
        
        AIEngine ai = new AIEngine(id, heuristic, depthCutOff);
        
        aiThread = new AIThread(ai);
        aiThread.start();
    }

    @Override
    public void postOpponentMove(PlayerID opponent, CellPosition move) {
        assert aiThread != null;
        aiThread.postMove(move);
    }
    
    @Override
    public void requestMove(MoveListener listener) {
        assert aiThread != null;
        super.requestMove(listener);
        aiThread.requestMove();
    }
    
    /**
     * Class implementing the thread for running the AI.
     * @author Michael Ekstrand <ekstrand@cs.umn.edu>
     */
    private class AIThread extends Thread {
        /** Queue of commands for the thread to handle */
        private BlockingQueue<Command> commands;
        /** Sequence of moves that has been made in this game */
        private List<CellPosition> moveSequence;
        int rows, columns, goal;
        /** The ID of the first player of the game */
        PlayerID firstPlayer;

        /** The active AI engine */
        private AIEngine aiEngine;

        /**
         * Construct a new AI thread to use a particular engine.
         * @param ai The AI engine to use.
         */
        public AIThread(AIEngine ai) {
            super("AI");
            aiEngine = ai;
            commands = new LinkedBlockingQueue<Command>();
            moveSequence = new ArrayList<CellPosition>();
            rows = game.getBoard().getRows();
            columns = game.getBoard().getColumns();
            goal = game.getBoard().getGoal();
            firstPlayer = game.getCurrentPlayer();
        }

        /**
         * Instruct the AI thread to shut down.
         */
        public void shutdown() {
            commands.add(new Command(Command.SHUTDOWN));
        }

        /**
         * Request a move from the AI thread.
         */
        public void requestMove() {
            commands.add(new Command(Command.MOVE));
        }

        /**
         * Post an opponent move to the AI thread.
         */
        public void postMove(CellPosition move) {
            commands.add(new Command(Command.POST, move));
        }

        /**
         * Master run method for the AI thread.  This implements the
         * command-handling loop over {@link #commands}.
         */
        @Override
        public synchronized void run() {
            logger.fine("AI thread beginning");
            try {
                boolean keepRunning = true;
                while (keepRunning) {
                    Command cmd = commands.take();
                    switch (cmd.id) {
                    case Command.SHUTDOWN:
                        logger.fine("AI thread requested to shutdown");
                        keepRunning = false;
                        break;
                    case Command.MOVE:
                        logger.fine("Move requested from AI thread");
                        handleMoveRequest();
                        break;
                    case Command.POST:
                        logger.fine("Move " + cmd.cell + " posted to AI thread");
                        moveSequence.add(cmd.cell);
                        break;
                    default:
                        throw new RuntimeException("Unknown loop command");
                    }
                }
            } catch (InterruptedException e) {
                logger.severe("Unexpected interruption");
                System.exit(3);
            }
            logger.fine("AI thread terminating");
        }

        /**
         * Handle a move request from the game.  This method asks the AI for a
         * move, and then posts the move back to the game in the event loop.
         */
        private void handleMoveRequest() {
            CellPosition move = null;
            TreeNode node = new TreeNode(firstPlayer, rows, columns, goal);
            for (CellPosition m: moveSequence)
                node = node.getChild(m);
            long moveStart = System.currentTimeMillis();
            try {
                move = aiEngine.getNextMove(node);
            } catch (RuntimeException e) {
                logger.log(Level.SEVERE,
                        "Error occured in AI move: " + e.getMessage(), e);
                System.exit(2);
            } catch (OutOfMemoryError err) {
                logger.severe("AI ran out of memory, terminating");
                System.exit(3);
            }
            long moveFinish = System.currentTimeMillis();
            double time = (moveFinish - moveStart) / 1000.0;
            logger.fine(String.format("AI moved to %s (took %.2fs)",
                                      move, time));
            moveSequence.add(move);
            invokeLater(new MoveNotifier(move));
        }

        /**
         * Encapsulation of the concept of an AI thread command.
         */
        private class Command {
            public static final int SHUTDOWN = -1;
            public static final int MOVE = 1;
            public static final int POST = 2;
            
            public final int id;
            public final CellPosition cell;
            
            public Command(int id) {
                this(id, null);
            }
            
            public Command(int id, CellPosition cell) {
                this.id = id;
                this.cell = cell;
            }
        }
    }

    /**
     * Class used to notify the game of a move.
     */
    private class MoveNotifier implements Runnable {
        private CellPosition move;
        public MoveNotifier(CellPosition move) {
            this.move = move;
        }
        
        public void run() {
            notifyMove(move);
        }
    }
}
