/* $Id: Game.java,v 1.5 2007/11/28 17:49:42 ekstrand Exp $ */

package csci1902.hw6.game;

import java.util.*;
import java.util.logging.Logger;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.SwingUtilities;

/**
 * Class encapsulating the game of circular tic-tac-toe.
 *
 * This class, in conjunction with {@link Board}, implements the core game
 * functionality - alternating players, detecting game-over, etc.
 *
 * @see Board
 * @see Player
 *
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 */
public class Game {
    private static Random random = new Random();
    private static Logger logger = Logger.getLogger("csci1902.hw6.Game");
    
    private Player player1, player2;
    private List<ChangeListener> listeners = new LinkedList<ChangeListener>();
    private Player currentPlayer;
    private Player otherPlayer;
    private int rows, columns, goal;
    private Board board;

    /**
     * Construct a new default game (4x8 grid, goal 4) with the specified
     * players.
     */
    public Game(Player player1, Player player2) {
        this(4, 8, 4, player1, player2);
    }

    /**
     * Construct a new game with a specified size, goal and players.  Note
     * that player1/player2 makes no difference to game order; the game will
     * randomly select a player to go first.
     * 
     * @param rows The number of rows in the game board.
     * @param cols The number of columns in the game board.
     * @param goal The goal of the game (# in a row).
     * @param player1 The "first" player.
     * @param player2 The other player.
     */
    public Game(int rows, int cols, int goal, Player player1, Player player2) {
        this.player1 = player1;
        this.player2 = player2;
        this.rows = rows;
        this.columns = cols;
        this.goal = goal;
        newGame();
    }
    
    /**
     * Obtain the game's random number generator.
     * @return The random number generator to use.
     */
    public static Random getRandom() {
        return random;
    }

    /**
     * Sets up a new game.
     */
    private void newGame() {
        if (random.nextBoolean()) {
            currentPlayer = player1;
            otherPlayer = player2;
        } else {
            currentPlayer = player2;
            otherPlayer = player1;
        }
        board = new Board(rows, columns, goal);
        player1.setGame(this);
        player2.setGame(this);
        notifyChanged();
    }

    /**
     * Start the game playing (ask the first player to make a move).
     */
    public void start() {
        currentPlayer.requestMove(new MoveResponder(currentPlayer));
    }

    /**
     * Handles a move by the specified player to a position.
     * @param player The player moving.
     * @param pos The position moved to.
     */
    private void handleMove(PlayerID player, CellPosition pos) {
        logger.fine("Received move to " + pos + " from " + player);
        if (!player.equals(currentPlayer.getId()))
            throw new BadMoveException("Player moved when not its turn");
        
        board = new Board(board, currentPlayer.getId(), pos);
        otherPlayer.postOpponentMove(currentPlayer.getId(), pos);
        switchPlayers();
        notifyChanged();
        if (!board.isFinished()) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    currentPlayer.requestMove(new MoveResponder(currentPlayer));
                }
            });
        }
    }

    /**
     * Get the current player.  At the start of the game, this will be the
     * first player.
     * @return The current player.
     */
    public PlayerID getCurrentPlayer() {
        return currentPlayer.getId();
    }

    /**
     * Get the game board.
     * @return The current game board.
     */
    public Board getBoard() {
        return board;
    }

    /** Switches players. */
    private void switchPlayers() {
        Player tmp = currentPlayer;
        currentPlayer = otherPlayer;
        otherPlayer = tmp;
    }

    /**
     * Register a listener interested in notifications of game state changes.
     * @param l The listener to add.  Its {@link
     * javax.swing.event.ChangeListener#stateChanged(ChangeEvent)} method will
     * be invoked whenever the game state changes.
     */
    public void addChangeListener(ChangeListener l) {
        listeners.add(l);
    }
    /**
     * Remove a listener from the change listener list.
     * @param l The listener to remove.
     */
    public void removeChangeListener(ChangeListener l) {
        listeners.remove(l);
    }
    /**
     * Notify all listeners of a change in game state.
     */
    protected void notifyChanged() {
        ChangeEvent evt = new ChangeEvent(this);
        for (ChangeListener l: listeners) {
            l.stateChanged(evt);
        }
    }

    /**
     * Class used to listen for moves from players.
     */
    private class MoveResponder implements MoveListener {
        private Player player;
        private boolean called = false;

        /**
         * Construct a new move listener.
         * @param player The player expected to move.
         */
        public MoveResponder(Player player) {
            this.player = player;
        }

        /**
         * Records a move.
         * @param pos The position moved to.
         */
        public void move(CellPosition pos) {
            if (called) 
                throw new RuntimeException(
                        "Cannot post multiple moves to the same listener");
            called = true;
            handleMove(player.getId(), pos);
        }
    }
}
